##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from .client import Client
from pyds8k.utils import dictionarize
from pyds8k.exceptions import NotFound
from logging import getLogger
from pyds8k import PYDS8K_DEFAULT_LOGGER

logger = getLogger(PYDS8K_DEFAULT_LOGGER)


class SCClient(object):
    """
    SC side client. Used to interaction with current side client.

    !--important: the id field of all resources are case insensitive--!
    """

    def __init__(self, service_address, user, password,
                 port=None,
                 hostname='',
                 ):
        self.client = Client(service_address, user, password,
                             port=port,
                             hostname=hostname,
                             )

    @dictionarize
    def get_system(self):
        return self.client.get_system()

    @dictionarize
    def get_volume(self, volume_id):
        return self.client.get_volumes(volume_id)

    @dictionarize
    def get_host(self, host_name):
        return self.client.get_hosts(host_name)

    @dictionarize
    def get_extentpool(self, pool_id):
        pool = self.client.get_pools(pool_id)
        try:
            # lazy-loading
            eserep = pool.eserep[0]
            pool.representation['eserep'] = eserep.representation
        except AttributeError:
            pass
        return pool

    @dictionarize
    def list_extentpools(self):
        """
        return extent pool list without ese capacity info.
        """
        return self.client.get_pools()

    @dictionarize
    def list_hosts(self):
        return self.client.get_hosts()

    @dictionarize
    def list_extentpool_volumes(self, pool_id):
        # two requests
        # pool = self.client.get_pools(pool_id)
        # return pool.volumes

        # one request
        return self.client.get_volumes_by_pool(pool_id)

    @dictionarize
    def list_extentpool_virtualpool(self, pool_id):
        return self.client.get_eserep_by_pool(pool_id)

    @dictionarize
    def list_flashcopies(self):
        return self.client.get_flashcopies()

    @dictionarize
    def list_volume_flashcopies(self, volume_id):
        # two requests
        # volume = self.client.get_volumes(volume_id)
        # return volume.flashcopy

        # one request
        return self.client.get_flashcopies_by_volume(volume_id)

    @dictionarize
    def list_remotecopies(self):
        return self.client.get_pprc()

    @dictionarize
    def list_volume_remotecopies(self, volume_id):
        return self.client.get_pprc_by_volume(volume_id)

    @dictionarize
    def list_logical_subsystems(self):
        return self.client.get_lss()

    @dictionarize
    def list_lss_volumes(self, lss_number):
        return self.client.get_volumes_by_lss(lss_number)

    @dictionarize
    def list_fcports(self):
        return self.client.get_ioports()

    def list_ioenclosures(self):
        pass

    def get_client_descriptor(self):
        pass

    def list_known_wwpns(self):
        pass

    def get_used_lun_numbers_by_host(self, host_name):
        mappings = self.client.get_mappings_by_host(host_name)
        return [mapping.id for mapping in mappings]

    def create_volumes(self, pool_id, capacity_in_GiB, sam,
                       volume_names_list):
        return self.client.create_volumes(name_col=volume_names_list,
                                          cap=capacity_in_GiB,
                                          pool=pool_id,
                                          tp=sam)

    def rename_volume(self, volume_id, new_name):
        return self.client.update_volume_rename(volume_id=volume_id,
                                                new_name=new_name)

    def extend_volume(self, volume_id, new_size_in_GiB):
        return self.client.update_volume_extend(volume_id=volume_id,
                                                new_size=new_size_in_GiB,
                                                captype='gib')

    def delete_volume(self, volume_id):
        # remember to unmap all hosts before delete.
        return self.client.delete_volume(volume_id=volume_id)

    def relocate_volume(self, volume_id, new_pool_id):
        return self.client.update_volume_move(volume_id=volume_id,
                                              new_pool=new_pool_id)

    def create_extentpool_virtualpool(self):
        pass

    def remove_extentpool_virtualpool(self):
        pass

    def crate_host(self, host_name, wwpn, host_type='VMware'):
        hosts = self.client.create_host(host_name=host_name,
                                        hosttype=host_type
                                        )
        self.attach_hostport_to_host(host_name=hosts[0].id, wwpn=wwpn)
        return hosts[0].id

    def delete_host(self, host_name):
        # delete a host will delete all the attached host ports.
        return self.client.delete_host(host_name=host_name)

    def attach_hostport_to_host(self, host_name, wwpn):
        return self._get_attach_or_create_host_port(host_name=host_name,
                                                    wwpn=wwpn
                                                    )

    def detach_hostport_from_host(self, wwpn):
        return self.client.delete_host_port(port_id=wwpn)

    def map_volume_to_host(self, host_name, volume_id, lunid):
        return self.client.map_volume_to_host(host_name=host_name,
                                              volume_id=volume_id,
                                              lunid=lunid
                                              )

    def unmap_volume_from_host(self, host_name, lunid):
        return self.client.unmap_volume_from_host(host_name=host_name,
                                                  lunid=lunid
                                                  )

    def _get_attach_or_create_host_port(self, host_name, wwpn):
        try:
            host_port = self.client.get_host_port(port_id=wwpn)
            return self.client.update_host_port_change_host(
                port_id=host_port.id,
                host_name=host_name
            )
        except NotFound:
            logger.debug(
                'host port {} is not found, creating new...'.format(wwpn)
            )
            return self.client.create_host_port(port_id=wwpn,
                                                host_name=host_name
                                                )

    # APIs below are deprecated.

    # def list_cached_systems(self):
    #    pass

    def list_volume_groups(self):
        pass

    def get_volume_group(self):
        pass

    def attach_volume_group(self):
        pass

    def detach_volume_group(self):
        pass

    def add_volumes_to_volgrp(self):
        pass

    def remove_volume_from_volgrp(self):
        pass

    def create_volume_group(self):
        pass

    def remove_volume_group(self):
        pass

    def list_scsi_host_ports(self):
        pass

    def get_scsi_host_port(self):
        pass

    def remove_scsi_host_port(self):
        pass

    def create_scsi_host_port(self):
        pass



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Host Volume Mapping interface.
"""

from .common.base import Base, BaseManager
from .common.types import DS8K_VOLMAP
from .volumes import Volume, VolumeManager


class Volmap(Base):
    id_field = 'lunid'
    _template = {'lunid': '',
                 'volume': '',
                 }
    related_resource = {'_volume': (Volume, VolumeManager),
                        }

    def __repr__(self):
        return "<Host Volume Mapping: {0}>".format(self._get_id())


class VolmapManager(BaseManager):
    """
    Manage Host Volume Mapping resources.
    """
    resource_class = Volmap
    resource_type = DS8K_VOLMAP

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id, url=url,
                         obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)


RESOURCE_TUPLE = (Volmap, VolmapManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import os
from importlib import import_module

_PATH = os.path.abspath(os.path.dirname(__file__))
mocks = set([os.path.splitext(resource)[0]
             for resource in os.listdir(_PATH)
             if os.path.isfile(os.path.join(_PATH, resource)) and
             not str(resource).startswith('__init__')
             ])
success_response_one = {}
success_response_all = {}

for re in mocks:
    success_response_one[re] = import_module(
        '{0}.{1}'.format(__name__, re)
    ).ONE
    success_response_all[re] = import_module(
        '{0}.{1}'.format(__name__, re)
    ).ALL



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Extent pool interface.
"""

from .common.types import DS8K_POOL, DS8K_VOLUME, \
    DS8K_TSEREP, DS8K_ESEREP
from .common.base import Base, ReadOnlyManager
from .common.mixins import VolumeMixin
from pyds8k.exceptions import IDMissingError


# Note: VolumeMixin will override the methods with
# same name in RootResourceMixin.
class Pool(VolumeMixin, Base):
    # id_field = 'id'
    _template = {'id': '',
                 'name': '',
                 'node': '',
                 'stgtype': '',
                 'cap': '',
                 'capalloc': '',
                 'capavail': '',
                 'overprovisioned': '',
                 'easytier': '',
                 'tieralloc': [],
                 'threshold': '',
                 'real_capacity_allocated_on_ese': None,
                 'virtual_capacity_allocated_on_ese': None,
                 DS8K_VOLUME: '',
                 DS8K_TSEREP: '',
                 DS8K_ESEREP: '',
                 }
    related_resources_collection = (DS8K_VOLUME, DS8K_TSEREP, DS8K_ESEREP)

    def __repr__(self):
        return "<Extent Pool: {0}>".format(self._get_id())

    def get_tserep(self):
        if not self.id:
            raise IDMissingError()
        tserep = self.all(DS8K_TSEREP).list()
        self._start_updating()
        setattr(self, DS8K_TSEREP, tserep)
        self._stop_updating()
        return tserep

    def get_TSE_rep(self):
        return self.get_tserep()[0]

    def get_eserep(self):
        if not self.id:
            raise IDMissingError()
        eserep = self.all(DS8K_ESEREP).list()
        self._start_updating()
        setattr(self, DS8K_ESEREP, eserep)
        self._stop_updating()
        return eserep

    def get_ESE_rep(self):
        return self.get_eserep()[0]

    def delete_tserep(self):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_TSEREP).delete()

    def delete_eserep(self):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_ESEREP).delete()

    def update_tserep_cap(self, cap, captype=''):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_TSEREP
                        ).update({'cap': cap, 'captype': captype})

    def update_eserep_cap(self, cap, captype=''):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_ESEREP
                        ).update({'cap': cap, 'captype': captype})

    def update_tserep_threshold(self, threshold):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_TSEREP
                        ).update({'threshold': threshold})

    def update_eserep_threshold(self, threshold):
        if not self.id:
            raise IDMissingError()
        return self.all(DS8K_ESEREP
                        ).update({'threshold': threshold})


class PoolManager(ReadOnlyManager):
    """
    Manage Extent Pool resources.
    """
    resource_class = Pool
    resource_type = DS8K_POOL


RESOURCE_TUPLE = (Pool, PoolManager)



"""
Usage examples of the DS8000 Python Client.
"""

from pyds8k.client.ds8k.v1.client import Client

# Connect storage array through SSH
restclient = Client(service_address='ip_address', user='username',
                    password='password')

# Create volumes
restclient.create_volumes(name_col=["volume_name"],
                          cap="capacity_in_GiB",
                          pool="pool_id",
                          tp="none")

# Get Volume from storage
vol = restclient.get_volumes('vol_id')
print(vol['name'])

# Get all the volume from storage by lss
lss_vols = restclient.get_volumes_by_lss('lss_number')
for vol in lss_vols:
    print(vol['name'])

# Get all the volumes from storage
vols = restclient.get_volumes()
for vol in vols:
    print(vol['name'])

# Get Pool from storage
pool = restclient.get_pools('pool_id')
print(pool['name'])
print(pool.eserep[0])

# Getting all of the pools from storage
pools = restclient.get_pools()
for p in pools:
    print(p['name'])



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import httpretty
from datetime import datetime
from .base import TestDS8KWithConnect
from ...data import get_response_list_json_by_type
from pyds8k.resources.ds8k.v1.common.types import DS8K_EVENT
from pyds8k.exceptions import InvalidArgumentError

event_list_response = get_response_list_json_by_type(DS8K_EVENT)


class TestHost(TestDS8KWithConnect):

    @httpretty.activate
    def test_get_events_by_filter_set_severity(self):
        url = '/events'

        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=event_list_response,
                               content_type='application/json',
                               )
        self.system.get_events_by_filter(warning=True, error=True)
        req = httpretty.last_request()
        self.assertIsNotNone(req.querystring)
        self.assertIn('severity', req.querystring)
        self.assertEqual('warning,error', req.querystring.get('severity')[0])

    @httpretty.activate
    def test_get_events_by_filter_set_date_error(self):
        url = '/events'

        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=event_list_response,
                               content_type='application/json',
                               )
        with self.assertRaises(InvalidArgumentError):
            self.system.get_events_by_filter(before='test')

    @httpretty.activate
    def test_get_events_by_filter_set_date(self):
        url = '/events'
        before = datetime(2015, 4, 1)
        after = datetime(2015, 1, 1)

        httpretty.register_uri(httpretty.GET,
                               self.domain + self.base_url + url,
                               body=event_list_response,
                               content_type='application/json',
                               )
        self.system.get_events_by_filter(before=before, after=after)
        req = httpretty.last_request()
        self.assertIsNotNone(req.querystring)
        self.assertIn('before', req.querystring)
        self.assertIn('after', req.querystring)

        # httpretty unquote "+" and " " in a wrong way,
        # so I can not verify time zone here.
        self.assertEqual('2015-04-01T00:00:00',
                         req.querystring.get('before')[0][:-5]
                         )
        self.assertEqual('2015-01-01T00:00:00',
                         req.querystring.get('after')[0][:-5]
                         )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
DS8K resources base interface.
"""
from logging import getLogger
from pyds8k.messages import INVALID_TYPE
from pyds8k import PYDS8K_DEFAULT_LOGGER
from pyds8k.base import Resource, Manager
from .mixins import RootResourceMixin
from pyds8k.exceptions import OperationNotAllowed, \
    URLNotSpecifiedError, \
    FieldReadOnly
from ....utils import get_resource_class_by_name

logger = getLogger(PYDS8K_DEFAULT_LOGGER)


class Base(RootResourceMixin, Resource):
    # If there is a field named "id" in response data,
    # the id_field can't be set to value other than "id"
    id_field = 'id'
    url_field = 'link'
    base_url = '/api/v1'
    create_method = 'posta'
    # Required only in writable resources, fileds are from _template
    # Resource id is exclude.
    readonly_fileds = ()
    # Not like related_resource, related_resources_list is not set during
    # loading, its keys use lazy-loading to get details.
    related_resources_collection = ()

    def _add_details(self, info, force=False):
        super(Base, self)._add_details(info, force=force)
        self._start_updating()
        self._set_related_resources_collection()
        self._stop_updating()

    def _set_related_resources_collection(self):
        for key in self.related_resources_collection:
            res = self.representation.get(key)
            # If the related resources(should be a list) are not in info,
            # will empty them and wait for lazy-loading.
            if not isinstance(res, list):
                self.representation[key] = ''
                try:
                    delattr(self, key)
                except AttributeError:
                    pass
            # If the related resources(should be a list) are in info, set it.
            else:
                re_class, re_manager = self._get_resource_class_by_name(key)
                res_list = [re_class(self.client,
                                     manager=re_manager(self.client),
                                     info=r)
                            for r in res]
                setattr(self, key, res_list)

    def __setattr__(self, key, value):
        if key in self.readonly_fileds and not self.is_updating():
            raise FieldReadOnly(key)
        super(Base, self).__setattr__(key, value)
        try:
            if key in self.related_resources_collection:
                ids = [getattr(item, item.id_field) for item in value]
                self.representation[key] = ids
                if not self.is_updating():
                    self._set_modified_info_dict(key, ids)
        except AttributeError:
            pass

    def __getattr__(self, key):
        if key in self.related_resources_collection:
            try:
                return getattr(self, 'get_{}'.format(key))()
            except Exception as e:
                logger.debug(
                    "Can not get {} from {}, reason is: {}".format(
                        key, self, type(e)
                        )
                )
                raise AttributeError(key)
        return super(Base, self).__getattr__(key)

    def __repr__(self):
        return "<{0}: {1}>".format(self.__class__.__name__, self._get_id())

    def _get_resource_class_by_name(self, resource_type):
        prefix = '{}.{}'.format(self.client.service_type,
                                self.client.service_version
                                )
        return get_resource_class_by_name(resource_type, prefix)

    def _verify_type(self, new_type, valid_type_list):
        if new_type and not (new_type in valid_type_list):
            raise ValueError(
                INVALID_TYPE.format(', '.join(valid_type_list))
            )


class SingletonBase(Base):
    # A singleton resource has no id field by default
    id_field = '*'


class BaseManager(Manager):
    resource_class = Base
    response_key = 'data'
    resource_type = ''

    def _post(self, url='', body=None):
        post_body = None
        if not body:
            if self.managed_object is not None:
                post_body = self.managed_object._get_modified_info_dict()
                # repre = self.managed_object.representation
                # post_body = {key: value
                #             for key, value in repre.iteritems()
                #             if key not in self.managed_object.readonly_fileds
                #             }
            else:
                raise URLNotSpecifiedError()
        else:
            post_body = body
        return super(BaseManager, self)._post(url=url, body=post_body)

    # DS8K will use PUT in PATCH way, and don't use PATCH.
    def _put(self, url='', body=None):
        put_body = None
        if not url:
            if self.managed_object is not None:
                self.url = self.managed_object.url
                # use modified info here
                put_body = body if body else \
                    self.managed_object._get_modified_info_dict()
            else:
                raise URLNotSpecifiedError()
        else:
            self.url = url
            put_body = body
        resp, body = self.client.put(self.url,
                                     body=self._get_request_data(put_body)
                                     )
        data = self._get_data(body, method='PUT', response=resp)
        return resp, data

    def _patch(self, url='', body=None):
        return self._put(url=url, body=body)

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        raise OperationNotAllowed('get', self.resource_class.__name__)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        raise OperationNotAllowed('list', self.resource_class.__name__)

    def post(self, url='', body=None):
        raise OperationNotAllowed('post', self.resource_class.__name__)

    def posta(self, url='', body=None):
        raise OperationNotAllowed('posta', self.resource_class.__name__)

    def put(self, url='', body=None):
        raise OperationNotAllowed('put', self.resource_class.__name__)

    def patch(self, url='', body=None):
        raise OperationNotAllowed('patch', self.resource_class.__name__)

    def delete(self, url=''):
        raise OperationNotAllowed('delete', self.resource_class.__name__)


class ReadOnlyManager(BaseManager):

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id,
                         url=url, obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)


class SingletonBaseManager(BaseManager):

    def get(self, url='', obj_class=None, **kwargs):
        return self._get(url=url, obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from .. import base
from pyds8k.resources import utils


class TestUtils(base.TestCaseWithoutConnect):

    def test_update_resource_id_in_url(self):
        old_id = '1'
        new_id = '2'
        old_url_str = '/default/{}'.format(old_id)
        new_url_str = '/default/{}'.format(new_id)
        old_url_dict = {
                        'rel': 'self',
                        'href': old_url_str
                        }
        new_url_dict = {
                        'rel': 'self',
                        'href': new_url_str
                        }
        self.assertEqual(
                utils.update_resource_id_in_url(old_id, new_id, old_url_str),
                new_url_str
                         )
        self.assertEqual(
                         utils.update_resource_id_in_url(old_id, new_id,
                                                         old_url_dict, 'href'
                                                         ),
                         new_url_dict
                         )



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.base import Resource, DefaultManager
from . import base
import httpretty
import json
from pyds8k.messages import DEFAULT_SUCCESS_BODY_DICT
from .data import get_response_list_json_by_type, \
    get_response_list_data_by_type, \
    get_response_data_by_type, \
    get_response_json_by_type
from .data import action_response, action_response_json
from .data import default_template

info = {'id': 'v1', 'name': 'vol1'}

custom_method_get = {'msg': 'custom_method_get'}
custom_method_get_json = json.dumps(custom_method_get)

DEFAULT = 'default'
default_a_response = get_response_data_by_type(DEFAULT)
default_a_response_json = get_response_json_by_type(DEFAULT)
default_list_response = get_response_list_data_by_type(DEFAULT)
default_list_response_json = get_response_list_json_by_type(DEFAULT)


# Note: The ds8k's data parser will be treated as the default parser here.
class TestResource(base.TestCaseWithConnect):

    def setUp(self):
        super(TestResource, self).setUp()
        self.base_url = ''

    def test_one_all(self):
        url1 = '/default/a/default/b/default'
        url2 = '/default/a/default/b/default/c'
        vol1 = self.resource.one(
                                 DEFAULT,
                                 'a'
                                 ).one(
                                       DEFAULT,
                                       'b'
                                       ).all(DEFAULT)
        vol2 = self.resource.one(
                                 DEFAULT,
                                 'a'
                                 ).one(
                                       DEFAULT,
                                       'b'
                                       ).one(DEFAULT, 'c')

        # test rebuild url
        vol3 = vol2.one(
                        DEFAULT,
                        'a',
                        rebuild_url=True
                        ).one(
                              DEFAULT,
                              'b'
                              ).all(DEFAULT)

        self.assertIsInstance(vol1, Resource)
        self.assertIsInstance(vol2, Resource)
        self.assertIsInstance(vol3, Resource)

        self.assertIsInstance(vol1.parent, Resource)
        self.assertIsInstance(vol2.parent, Resource)
        self.assertIsInstance(vol3.parent, Resource)

        self.assertEqual(vol1.url, self.base_url + url1)
        self.assertEqual(vol2.url, self.base_url + url2)
        self.assertEqual(vol3.url, self.base_url + url1)

    @httpretty.activate
    def test_toUrl(self):
        domain = self.client.domain
        url = '/default/a/default/b/default/c'
        method = 'attach'
        body = {'test': 'test'}
        httpretty.register_uri(httpretty.GET,
                               domain + self.base_url + url + '/' + method,
                               body=custom_method_get_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.POST,
                               domain + self.base_url + url + '/' + method,
                               body=action_response_json,
                               content_type='application/json')
        vol = self.resource.one(
                                DEFAULT,
                                'a'
                                ).one(
                                      DEFAULT,
                                      'b'
                                      ).one(DEFAULT, 'c')
        _, body1 = vol.toUrl(method)
        self.assertEqual(vol.url, self.base_url + url)
        self.assertEqual(body1, custom_method_get)

        _, body2 = vol.toUrl(method, body)
        self.assertEqual(vol.url, self.base_url + url)
        self.assertEqual(body2, action_response['server'])

    @httpretty.activate
    def test_create_from_template_and_save(self):
        domain = self.client.domain
        url = '/default/a/default/b/default'
        httpretty.register_uri(
            httpretty.POST,
            domain + self.base_url + url,
            responses=[
                httpretty.Response(
                    body=action_response_json,
                    content_type='application/json',
                    adding_headers={
                        'Location': self.base_url + url + '/vol1_id'
                    },
                    status=201),
                httpretty.Response(
                    body=action_response_json,
                    content_type='application/json',
                    adding_headers={
                        'Location': self.base_url + url + '/vol2_id'
                    },
                    status=201),
                httpretty.Response(
                    body=action_response_json,
                    content_type='application/json',
                    adding_headers={
                        'Location': self.base_url + url + '/vol3_id'
                    },
                    status=201),
            ]
        )
        httpretty.register_uri(
            httpretty.PUT,
            domain + self.base_url + url + '/vol3_id',
            responses=[
                httpretty.Response(
                    body=action_response_json,
                    content_type='application/json',
                    adding_headers={
                        'Location': self.base_url + url + '/vol3_id'
                    },
                    status=201),
            ]
        )
        vol1 = self.resource.one(
            DEFAULT,
            'a'
            ).one(DEFAULT,
                  'b'
                  ).all(DEFAULT).create_from_template(default_template)
        self.assertIsInstance(vol1, Resource)
        self.assertIsInstance(vol1.manager, DefaultManager)
        self.assertEqual(vol1.name, default_template['name'])
        self.assertEqual(vol1.url, self.base_url + url)
        self.assertEqual(vol1.representation, default_template)
        resp1, data1 = vol1.save()
        self.assertIsInstance(data1[0], Resource)
        self.assertEqual(resp1.status_code, 201)
        self.assertEqual(resp1.headers['Location'],
                         self.base_url + url + '/vol1_id'
                         )
        self.assertEqual(resp1.headers['Location'], vol1.url)

        vol2 = self.resource.one(
                                 DEFAULT,
                                 'a'
                                 ).one(
                                       DEFAULT,
                                       'b'
                                       ).one(
                                             DEFAULT,
                                             'c'
                                       ).create_from_template(default_template)
        vol2._template = default_template
        vol2.name = 'vol2'
        self.assertIsInstance(vol2, Resource)
        self.assertIsInstance(vol2.manager, DefaultManager)
        self.assertEqual(vol2.name, 'vol2')
        self.assertEqual(vol2.url, self.base_url + url)
        rep = default_template.copy()
        rep.update({'name': 'vol2'})
        self.assertEqual(vol2.representation, rep)
        resp2, data2 = vol2.save()
        self.assertIsInstance(data2[0], Resource)
        self.assertEqual(resp2.status_code, 201)
        self.assertEqual(resp2.headers['Location'],
                         self.base_url + url + '/vol2_id'
                         )
        self.assertEqual(resp2.headers['Location'], vol2.url)

        rep_with_id = default_template.copy()
        rep_with_id.update({'name': 'vol3', 'id': 'vol3_id'})
        vol3 = self.resource.one(
            DEFAULT,
            'a'
            ).one(DEFAULT,
                  'b'
                  ).one(DEFAULT,
                        'c'
                        ).create_from_template(rep_with_id)
        self.assertIsInstance(vol3, Resource)
        self.assertIsInstance(vol3.manager, DefaultManager)
        self.assertEqual(vol3.name, 'vol3')
        self.assertEqual(vol3.id, 'vol3_id')
        self.assertEqual(vol3.url, self.base_url + url + '/vol3_id')
        self.assertEqual(vol3.representation, rep_with_id)
        resp3, data3 = vol3.save()
        # default create method is put if id is specified.
        self.assertEqual(data3, action_response.get('server'))
        self.assertEqual(resp3.status_code, 201)
        self.assertEqual(resp3.headers['Location'],
                         self.base_url + url + '/vol3_id'
                         )

    def test_create(self):
        pass

    @httpretty.activate
    def test_lazy_loading(self):
        domain = self.client.domain
        url_list = '/default'
        vol_id = default_a_response['data']['default'][0]['id']
        url_a = '/default/{}'.format(vol_id)
        httpretty.register_uri(httpretty.GET,
                               domain + self.base_url + url_list,
                               body=default_list_response_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.GET,
                               domain + self.base_url + url_a,
                               body=default_a_response_json,
                               content_type='application/json')

        de_list = self.resource.all(DEFAULT).list()
        de0 = de_list[0]
        self.assertIsInstance(de0, Resource)
        self.assertIsInstance(de0.manager, DefaultManager)
        de0._template = {'id': '', 'name': ''}
        self.assertEqual(
                         de0.id,
                         default_list_response['data']['default'][0]['id']
                         )
        self.assertFalse('name' in de0.representation)
        # 'unknown' is not in _template
        self.assertRaises(AttributeError, getattr, de0, 'unknown')
        self.assertFalse(de0.is_loaded())
        # loading details
        self.assertEqual(
                         de0.name,
                         default_a_response['data']['default'][0]['name']
                         )
        self.assertTrue('name' in de0.representation)
        self.assertTrue(de0.is_loaded())

    def test_get_url(self):
        self.assertEqual(self.resource._get_url('/test'), '/test')
        self.assertEqual(self.resource._get_url(
                                        {'rel': 'self', 'href': '/test'}),
                         '/test'
                         )
        self.assertEqual(self.resource._get_url(
                                    [{'rel': 'self', 'href': '/test'},
                                     {'rel': 'bookmark',
                                      'href': '/bookmark'}, ]),
                         '/test'
                         )
        self.assertEqual(self.resource._get_url(
                                    [{'rel': 'self_', 'href': '/test'},
                                     {'rel': 'bookmark',
                                      'href': '/bookmark'}, ]),
                         ''
                         )
        self.assertEqual(self.resource._get_url(
                                    [{'rel': 'self', 'href_': '/test'},
                                     {'rel': 'bookmark',
                                      'href_': '/bookmark'}, ]),
                         ''
                         )
        self.assertEqual(self.resource._get_url(
                                    [{'rel_': 'self', 'href': '/test'},
                                     {'rel_': 'bookmark',
                                      'href': '/bookmark'}, ]),
                         ''
                         )
        self.assertRaises(Exception, self.resource._get_url, object())

    def test_id(self):
        self.assertFalse(hasattr(self.resource, 'id'))
        self.assertFalse(hasattr(self.resource, '_id'))
        self.resource._add_details(info)
        self.assertTrue(hasattr(self.resource, 'id'))
        self.assertTrue(hasattr(self.resource, '_id'))

        def set_id(_id):
            self.resource.id = _id
        self.assertRaises(Exception, set_id, 'a')

    def test_modified_info_dict(self):
        re = Resource(self.client, DefaultManager(self.client))
        self.assertEqual(re._get_modified_info_dict(), {})
        re._set_modified_info_dict('key1', 'val1')
        self.assertEqual(re._get_modified_info_dict(), {'key1': 'val1'})
        re._del_modified_info_dict_key('key')
        self.assertEqual(re._get_modified_info_dict(), {'key1': 'val1'})
        re._del_modified_info_dict_key('key1')
        self.assertEqual(re._get_modified_info_dict(), {})
        re._set_modified_info_dict('key2', 'val2')
        self.assertEqual(re._get_modified_info_dict(), {'key2': 'val2'})
        re._del_modified_info_dict_keys({'key1': 'val1'})
        self.assertEqual(re._get_modified_info_dict(), {'key2': 'val2'})
        re._del_modified_info_dict_keys({'key2': 'val2'})
        self.assertEqual(re._get_modified_info_dict(), {})

        re1 = Resource(self.client,
                       DefaultManager(self.client),
                       )
        re1._template = {'key1': '', 'key2': ''}
        re1._add_details(info={'key1': 'val1'})
        self.assertEqual(re1._get_modified_info_dict(), {})
        self.assertEqual(re1.key1, 'val1')
        re1.key1 = 'val1_changed'
        self.assertEqual(
                         re1._get_modified_info_dict(),
                         {'key1': 'val1_changed'}
                         )
        self.assertEqual(re1.key1, 'val1_changed')

        # set attr not in _template
        re1.key3 = 'val3'
        self.assertEqual(
                         re1._get_modified_info_dict(),
                         {'key1': 'val1_changed'}
                         )

    def test_force_get(self):
        re1 = Resource(self.client,
                       DefaultManager(self.client),
                       )
        re1._template = {'key1': '', 'key2': ''}
        re1._add_details(info={'key1': 'val1'})
        self.assertEqual(re1._get_modified_info_dict(), {})
        self.assertEqual(re1.key1, 'val1')
        re1.key1 = 'val1_changed'
        self.assertEqual(
                         re1._get_modified_info_dict(),
                         {'key1': 'val1_changed'}
                         )
        self.assertEqual(re1.key1, 'val1_changed')

        re1._add_details(info={'key1': 'val1'})
        self.assertEqual(re1.key1, 'val1_changed')

        re1._add_details(info={'key1': 'val1'}, force=True)
        self.assertEqual(re1.key1, 'val1')

    @httpretty.activate
    def test_list(self):
        domain = self.client.domain
        url = '/default'
        url1 = default_a_response['data']['default'][0]['link']['href']
        httpretty.register_uri(httpretty.GET, domain + self.base_url + url,
                               body=default_list_response_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.GET, domain + url1,
                               body=default_a_response_json,
                               content_type='application/json')

        vol = self.resource.all(DEFAULT)
        self.assertEqual(vol.url, self.base_url + url)
        self.assertRaises(AttributeError, getattr, vol, 'id')
        vol_list = vol.list()
        self.assertIsInstance(vol_list, list)
        vol1 = vol_list[0]
        self.assertEqual(
                    vol1.url,
                    default_list_response['data']['default'][0]['link']['href']
                         )
        self.assertEqual(
                         vol1.id,
                         default_a_response['data']['default'][0]['id']
                         )

        # lazy loading
        vol1._template = {'id': '', 'name': ''}
        self.assertEqual(
                         vol1.name,
                         default_a_response['data']['default'][0]['name']
                         )

    @httpretty.activate
    def test_get(self):
        domain = self.client.domain
        url = default_a_response['data']['default'][0]['link']['href']
        vol_id = default_a_response['data']['default'][0]['id']
        httpretty.register_uri(httpretty.GET, domain + url,
                               body=default_a_response_json,
                               content_type='application/json')

        vol = self.resource.one(DEFAULT, vol_id)
        self.assertEqual(vol.url, url)
        self.assertEqual(vol.id, vol_id)
        vol.get()
        self.assertEqual(
                    vol.url,
                    default_a_response['data']['default'][0]['link']['href']
                         )
        self.assertEqual(
                         vol.name,
                         default_a_response['data']['default'][0]['name']
                         )

        vol1 = self.resource.all(DEFAULT).get(vol_id)
        self.assertEqual(
                    vol1.url,
                    default_a_response['data']['default'][0]['link']['href']
                         )
        self.assertEqual(
                         vol1.name,
                         default_a_response['data']['default'][0]['name']
                         )

    @httpretty.activate
    def test_post(self):
        # post append: tested in test_create_from_template_and_save

        # post: tested in test_toUrl
        pass

    @httpretty.activate
    def test_put(self):
        # put new: tested in test_create_from_template_and_save

        # put update:
        domain = self.client.domain
        url = default_a_response['data']['default'][0]['link']['href']
        vol_id = default_a_response['data']['default'][0]['id']
        httpretty.register_uri(httpretty.GET, domain + url,
                               body=default_a_response_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.PUT, domain + url,
                               body=json.dumps({'status': 'updated'}),
                               content_type='application/json',
                               status=200)

        vol = self.resource.one(DEFAULT, vol_id).get()
        self.assertEqual(vol.name,
                         default_a_response['data']['default'][0]['name']
                         )
        vol.name = 'vol1_rename'
        resp, data = vol.put()
        self.assertEqual(data, {'status': 'updated'})
        self.assertEqual(resp.status_code, 200)

    @httpretty.activate
    def test_patch(self):
        domain = self.client.domain
        url = default_a_response['data']['default'][0]['link']['href']
        vol_id = default_a_response['data']['default'][0]['id']
        httpretty.register_uri(httpretty.GET, domain + url,
                               body=default_a_response_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.PATCH, domain + url,
                               body=json.dumps({'status': 'updated'}),
                               content_type='application/json',
                               status=200)

        vol = self.resource.one(DEFAULT, vol_id).get()
        self.assertEqual(
                         vol.name,
                         default_a_response['data']['default'][0]['name']
                         )
        vol._template = default_template
        vol.name = 'vol1_rename_patch'
        self.assertEqual(
                         vol._get_modified_info_dict(),
                         {'name': 'vol1_rename_patch'}
                         )
        resp, data = vol.patch()
        self.assertEqual(data, {'status': 'updated'})
        self.assertEqual(resp.status_code, 200)

    @httpretty.activate
    def test_delete(self):
        domain = self.client.domain
        url = default_a_response['data']['default'][0]['link']['href']
        vol_id = default_a_response['data']['default'][0]['id']
        httpretty.register_uri(httpretty.GET, domain + url,
                               body=default_a_response_json,
                               content_type='application/json')
        httpretty.register_uri(httpretty.DELETE, domain + url,
                               content_type='application/json',
                               status=204)

        vol = self.resource.one(DEFAULT, vol_id).get()
        self.assertEqual(
                         vol.name,
                         default_a_response['data']['default'][0]['name']
                         )
        resp, data = vol.delete()
        self.assertEqual(resp.status_code, 204)
        self.assertEqual(data, DEFAULT_SUCCESS_BODY_DICT)

    def test_save(self):
        # save new: tested in test_create_from_template_and_save

        # save update: tested in test_put
        pass

    def test_update(self):
        # update put: tested in test_put

        # update patch: tested in test_patch
        pass

    def test_equal(self):
        re1 = Resource(self.client, resource_id='test')
        re2 = Resource(self.client, resource_id='test')
        re3 = Resource(self.client, resource_id='test3')
        self.assertTrue(re1 == re2)
        self.assertFalse(re1 == re3)
        self.assertFalse(re1 is re2)
        self.assertTrue(re1 in [re2])

    def test_update_list_field(self):
        re1 = Resource(self.client, resource_id='test')
        re1.re_list = [Resource(self.client, resource_id='test{}'.format(n))
                       for n in range(10)
                       ]
        re_not_in = Resource(self.client, resource_id='test11')
        re_in = Resource(self.client, resource_id='test1')
        with self.assertRaises(KeyError):
            re1._update_list_field('re_list', re_not_in, '-')
        with self.assertRaises(KeyError):
            re1._update_list_field('re_list', re_in)
        re1._update_list_field('re_list', re_not_in)
        self.assertTrue(re_not_in in re1.re_list)
        re1._update_list_field('re_list', re_in, '-')
        self.assertTrue(re_in not in re1.re_list)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

"""
Storage volume interface.
"""

from .common.base import Base, BaseManager
from .common import types
from .pools import Pool, PoolManager
from .lss import LSS, LSSManager
from .hosts import Host, HostManager


class Volume(Base):
    # id_field = 'id'

    # Set the value to None if the field is not required when creation.
    _template = {'id': None,
                 'name': '',
                 'state': '',
                 'cap': '',
                 'real_cap': None,
                 'virtual_cap': None,
                 'captype': '',
                 'stgtype': '',
                 'VOLSER': None,
                 'allocmethod': '',
                 'tp': '',
                 'capalloc': '',
                 'MTM': None,
                 'datatype': '',
                 'easytier': '',
                 'tieralloc': [],
                 'lss': '',
                 'pool': '',
                 types.DS8K_HOST: None,
                 types.DS8K_FLASHCOPY: None,
                 types.DS8K_PPRC: None,
                 }
    fb_template = _template.copy()
    fb_template.update({'stgtype': types.DS8K_VOLUME_TYPE_FB})

    ckd_template = _template.copy()
    ckd_template.update({'stgtype': types.DS8K_VOLUME_TYPE_FB})

    template_dict = {types.DS8K_VOLUME_TYPE_FB: fb_template,
                     types.DS8K_VOLUME_TYPE_CKD: ckd_template,
                     }
    readonly_fileds = ('state', 'stgtype', 'VOLSER', 'allocmethod', 'tp',
                       'capalloc', 'MTM', 'datatype', 'easytier', 'tieralloc',
                       'lss'
                       )
    related_resource = {'_pool': (Pool, PoolManager),
                        '_lss': (LSS, LSSManager)
                        }
    related_resources_collection = (types.DS8K_HOST,
                                    types.DS8K_FLASHCOPY,
                                    types.DS8K_PPRC)

    def __init__(self, client, manager=None, url='', info={},
                 resource_id=None,
                 parent=None,
                 loaded=False,
                 volume_type=types.DS8K_VOLUME_TYPE_FB,
                 ):
        super(Volume, self).__init__(client,
                                     manager=manager,
                                     url=url,
                                     info=info,
                                     resource_id=resource_id,
                                     parent=parent,
                                     loaded=loaded,
                                     )
        self.volume_type = volume_type
        self._verify_volume_type()
        self._template = self.get_template_from_volume_type(volume_type)

    def __repr__(self):
        return "<Storage Volume: {0}>".format(self._get_id())

    def get_template_from_volume_type(self, volume_type):
        return self.template_dict[volume_type]

    def _verify_volume_type(self):
        self._verify_type(self.volume_type, types.DS8K_VOLUME_TYPES)

    def _set_hosts(self):
        host_list = self.representation.get('host', [])
        if host_list:
            host_obj_list = []
            for host in host_list:
                host_obj = Host(self.client,
                                manager=HostManager(self.client),
                                info=host,
                                loaded=False,
                                )
                host_obj_list.append(host_obj)
            self.representation[types.DS8K_HOST] = [
                h.host_id for h in host_obj_list]
            setattr(self, types.DS8K_HOST, host_obj_list)

    def _add_details(self, info, force=False):
        super(Volume, self)._add_details(info, force=force)
        self._start_updating()
        self._set_hosts()
        self._stop_updating()


class VolumeManager(BaseManager):
    """
    Manage Storage Volume resources.
    """
    resource_class = Volume
    resource_type = types.DS8K_VOLUME

    def get(self, resource_id='', url='', obj_class=None, **kwargs):
        return self._get(resource_id=resource_id, url=url,
                         obj_class=obj_class, **kwargs)

    def list(self, url='', obj_class=None, body=None, **kwargs):
        return self._list(url=url, obj_class=obj_class, body=body, **kwargs)

    # def post(self, url='', body=None):
    #    return self._post(url, body)

    def posta(self, url='', body=None):
        return self._posta(url=url, body=body)

    def put(self, url='', body=None):
        return self._put(url=url, body=body)

    def patch(self, url='', body=None):
        return self._patch(url=url, body=body)

    def delete(self, url=''):
        return self._delete(url=url)


RESOURCE_TUPLE = (Volume, VolumeManager)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

import os
import time
import configparser
from importlib import import_module
from pyds8k.messages import GET_CONFIG_SETTINGS_IOERROR, \
    GET_CONFIG_SETTINGS_ERROR

_PATH = os.path.abspath(os.path.dirname(__file__))
CONFIG_FILE_NAME = 'config.ini'
CONFIG_FILE_PATH = os.path.join(_PATH, CONFIG_FILE_NAME)
logger = None

# HTTP STATUS CODES
HTTP200 = 200
HTTP204 = 204
HTTP404 = 404
HTTP500 = 500

# HTTP METHODS
POSTA = 'POST-to-Append'
POST = 'POST'
GET = 'GET'
PUT = 'PUT'
PATCH = 'PATCH'
DELETE = 'DELETE'


def _get_logger():
    from logging import getLogger
    from pyds8k import PYDS8K_DEFAULT_LOGGER
    global logger
    if not logger:
        logger = getLogger(PYDS8K_DEFAULT_LOGGER)
    return logger


def get_subclasses(cls):
    subclasses = cls.__subclasses__()
    for sub in list(subclasses):
        subclasses.extend(get_subclasses(sub))
    return subclasses


def get_config_settings(category="settings"):
    result_dict = dict()
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        for setting, value in config.items(category):
            result_dict[setting] = value
    except IOError as e:
        _get_logger().debug(GET_CONFIG_SETTINGS_IOERROR.format(
            CONFIG_FILE_PATH,
            str(e)
            )
        )
    except Exception as e:
        _get_logger().error(GET_CONFIG_SETTINGS_ERROR.format(str(e)))
    return result_dict


def get_config_all():
    result_dict = dict()
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        for section in config.sections():
            result_dict[section] = dict()
            for setting, value in config.items(section):
                result_dict[section][setting] = value
    except IOError as e:
        _get_logger().debug(GET_CONFIG_SETTINGS_IOERROR.format(
            CONFIG_FILE_PATH,
            str(e)
            )
        )
    except Exception as e:
        _get_logger().error(GET_CONFIG_SETTINGS_ERROR.format(str(e)))
    return result_dict


def get_config_all_items():
    result_dict = dict()
    try:
        config = configparser.ConfigParser()
        config.read(CONFIG_FILE_PATH)
        for section in config.sections():
            for setting, value in config.items(section):
                result_dict[setting] = value
    except IOError as e:
        _get_logger().debug(GET_CONFIG_SETTINGS_IOERROR.format(
            CONFIG_FILE_PATH,
            str(e)
            )
        )
    except Exception as e:
        _get_logger().error(GET_CONFIG_SETTINGS_ERROR.format(str(e)))
    return result_dict


def get_config_by_name(name):
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    return config.get('settings', name)


def set_config_by_name(name, value):
    config = configparser.ConfigParser()
    config.read(CONFIG_FILE_PATH)
    config.set('settings', name, value)

    with open(CONFIG_FILE_PATH, 'wb') as config_file:
        config.write(config_file)


"""
def get_default_service_type():
    return get_config_by_name('default_service_type')


def get_default_service_version():
    return get_config_by_name('default_service_version')


def get_runtime_service_type():
    return get_config_by_name('runtime_service_type')


def get_service_type():
    return get_runtime_service_type() or get_default_service_type()


def set_runtime_service_type(service_type):
    set_config_by_name('runtime_service_type', service_type)
"""


def get_request_parser_class(service_type):
    prefix = service_type
    Parser = import_module('{0}.dataParser.{1}'.format(__package__, prefix)
                           )
    return Parser.RequestParser


def get_response_parser_class(service_type):
    prefix = service_type
    Parser = import_module('{0}.dataParser.{1}'.format(__package__, prefix)
                           )
    return Parser.ResponseParser


def timer(func):
    def inner(self, *args, **kwargs):
        start = time.time()
        result = func(self, *args, **kwargs)
        end = time.time()
        _get_logger().info(
            "Successfully called method '{}' in {} seconds".format(
                func.__name__,
                round(end - start, 2)
            )
        )
        return result
    return inner


def res_timer_recorder(func):
    def inner(self, *args, **kwargs):
        start = time.time()
        res = func(self, *args, **kwargs)
        end = time.time()
        sec = round(end - start, 2)
        if not res:
            _get_logger().info(
                "Successfully got 0 resources in {} seconds".format(sec)
            )
            return []
        _get_logger().info(
            "Successfully got {} resources in {} seconds, \
{} seconds per 100 instances.".format(
                len(res),
                sec,
                round(sec / len(res) * 100, 2)
            )
        )
        return res
    return inner


def dictionarize(func):
    def inner(self, *args, **kwargs):
        res_obj = func(self, *args, **kwargs)
        if not isinstance(res_obj, list):
            res_obj = [res_obj, ]
        coverted = []
        for res in res_obj:
            coverted.append(res.representation)
        return coverted
    return inner


def is_absolute_url(url):
    if url.startswith('/'):
        return False
    elif '//' in url:
        return True
    # Don't verify the URI's validation here.
    else:
        return True



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from pyds8k.resources.ds8k.v1.common.types import DS8K_VOLUME
from pyds8k.dataParser.ds8k import RequestParser, ResponseParser
import json
from .. import base
from ..data import get_response_list_data_by_type, \
    get_response_data_by_type
from ..data import token_response_error, default_request

info = {'id': 'v1', 'name': 'vol1'}

custom_method_get = {'data': 'custom_method_get'}
custom_method_post = {'data': 'custom_method_post'}
custom_method_get_json = json.dumps(custom_method_get)
custom_method_post_json = json.dumps(custom_method_post)
volume_a_response = get_response_data_by_type(DS8K_VOLUME)
volume_list_response = get_response_list_data_by_type(DS8K_VOLUME)


class TestDataParser(base.TestCaseWithoutConnect):

    def test_responseParser(self):
        re = ResponseParser(volume_a_response, 'volumes')
        self.assertEqual(re.response_key, 'data')
        self.assertEqual(re.url_field, 'link')
        self.assertEqual(
                         re.get_representations(),
                         volume_a_response['data']['volumes']
                         )
        re.representation = re.get_representations()[0]
        self.assertEqual(
                        re.get_link(),
                        volume_a_response['data']['volumes'][0]['link']['href']
                         )

        re1 = ResponseParser(volume_list_response, 'volumes')
        self.assertEqual(re1.response_key, 'data')
        self.assertEqual(re1.url_field, 'link')
        self.assertEqual(
                         re1.get_representations(),
                         volume_list_response['data']['volumes']
                         )
        re1.representation = re1.get_representations()[0]
        self.assertEqual(
                    re1.get_link(),
                    volume_list_response['data']['volumes'][0]['link']['href']
                         )

        re2 = ResponseParser(token_response_error)
        self.assertEqual(
                         re2.get_status_body(),
                         token_response_error['server']
                         )
        self.assertEqual(
                         re2.get_error_code(),
                         token_response_error['server']['code']
                         )
        self.assertEqual(
                         re2.get_error_msg(),
                         token_response_error['server']['message']
                         )
        self.assertEqual(
                         re2.get_status(),
                         token_response_error['server']['status']
                         )

    def test_requestParser(self):
        re = RequestParser(default_request['request']['params'])
        self.assertEqual(re.request_key, 'request')
        self.assertEqual(re.get_request_data(), default_request)



##############################################################################
# Copyright 2019 IBM Corp.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
##############################################################################

from logging import getLogger
from pyds8k import PYDS8K_DEFAULT_LOGGER
from pyds8k.dataParser.base import BaseRequestParser, \
    BaseResponseParser
from pyds8k import messages
from pyds8k.exceptions import URLParseError, \
    RepresentationParseError, \
    IDMissingError

logger = getLogger(PYDS8K_DEFAULT_LOGGER)

success_status = ('ok', 'successful', )


class RequestParser(BaseRequestParser):
    """
    Parse the data user wants to send to server,
    in the right format that server defined.
    """
    request_key = 'request'
    request_key_multi = 'request'
    param_key = 'params'

    def __init__(self, raw_data, resource_key=''):
        if not (isinstance(raw_data, list) or isinstance(raw_data, dict)):
            raise TypeError(messages.NEED_A_DICT_OR_DICT_LIST.format(raw_data))
        self.raw_data = raw_data
        self.request_data = None

    def get_raw_data(self):
        return self.raw_data

    def get_request_data(self):
        data = {}
        if not isinstance(self.raw_data, list):
            data[self.request_key] = {self.param_key: {}}
            data[self.request_key][self.param_key].update(self.raw_data)
        else:
            data[self.request_key_multi] = {self.param_key: []}
            for d in self.raw_data:
                data[self.request_key_multi][self.param_key].append(d)
        self.request_data = data
        return data


class ResponseParser(BaseResponseParser):
    """
    Parse response data, to get resource link, representation, etc.
    """
    response_key = 'data'
    status_key = 'server'
    url_field = 'link'

    def __init__(self, raw_data, resource_key=''):
        self.raw_data = raw_data
        self.resource_key = resource_key
        self.representations = None
        self.representation = None
        self.status_body = None

    def get_raw_data(self):
        return self.raw_data

    def get_link(self, representation=None):
        rep = None
        if representation:
            rep = representation
        elif self.representation:
            rep = self.representation
        else:
            raise Exception(messages.REPRESENTATION_NOT_FOUND)
        return self.__class__.get_link_from_representation(rep)

    def get_representations(self):
        data = self.raw_data.get(self.response_key, self.raw_data)
        parsed = self._parse_data(data)
        self.representations = parsed
        return parsed

    def get_posta_response_data(self):
        MULTIFLAG = "responses"
        data = self.raw_data.get(MULTIFLAG, self.raw_data)
        if not isinstance(data, list):
            data = [data, ]
        res = []
        for s_data in data:
            res_status_body = s_data.get(self.status_key)
            status = self.get_status(res_status_body)
            # return data part if the status is ok.
            if str(status).lower() in success_status:
                res_data = s_data.get(self.response_key)
                res_url = self.get_link_from_representation(s_data)
                if not res_data:
                    res.append(({self.resource_data_key: None}, res_url))
                else:
                    res.append(
                        ({self.resource_data_key: self._parse_data(res_data)[0]},  # noqa
                         res_url
                         )
                    )
            # return status part if something failed.
            else:
                res.append(({self.error_status_key: res_status_body}, None))
        return res

    def get_status_body(self):
        self.status_body = self.raw_data.get(self.status_key, self.raw_data)
        return self.status_body

    def get_error_code(self, status_body=None):
        if not self.status_body:
            self.status_body = self.get_status_body()
        data = status_body or self.status_body
        return data.get('code', '')

    def get_error_msg(self, status_body=None):
        if not self.status_body:
            self.status_body = self.get_status_body()
        data = status_body or self.status_body
        return data.get('message', '')

    def get_status(self, status_body=None):
        if not self.status_body:
            self.status_body = self.get_status_body()
        data = status_body or self.status_body
        return data.get('status', '')

    @classmethod
    def get_link_from_representation(cls, representation):
        url_objects = representation.get(cls.url_field)
        if not url_objects:
            return ""
        url = cls._get_url(url_objects)
        return url

    @classmethod
    def get_resource_id_from_url(self, url, resource_type):
        if url.endswith('/'):
            url = url[:-1]
        url_frag = url.split('/')
        if len(url_frag) > 1 and url_frag[-2] == resource_type:
            return url_frag[-1]
        logger.debug("Failed to get resource id from url {}".format(url))
        raise IDMissingError()

    @classmethod
    def _get_url(cls, urls):
        if isinstance(urls, str):
            return urls
        elif isinstance(urls, dict):
            urls = [urls]
        elif isinstance(urls, list):
            pass
        else:
            raise URLParseError()
        for url in urls:
            if url.get('rel') == 'self':
                return url.get('href', '')
        return ''

    def _parse_data(self, data):
        parsed = None
        if isinstance(data, dict):
            if self.resource_key:
                parsed = data.get(self.resource_key)
                if parsed is None:
                    logger.debug(
                        "Failed to parse resource from data, return raw data: {}".format(data)  # noqa
                    )
                    parsed = data
        else:
            raise RepresentationParseError()
        if not isinstance(parsed, list):
            parsed = [parsed]
        return parsed
